/*
 * conffile.c
 *
 * This file is part of the msmtp package, a very simple programm that 
 * works as an "SMTP plugin" for Mutt and possibly other MUAs.
 *
 * Copyright (C) 2000, 2003	Martin Lambers <marlam@users.sourceforge.net>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <errno.h>
extern int errno;
#include <string.h>

#include "esmtp.h"
#include "list.h"
#include "xmalloc.h"
#include "merror.h"
#include "conffile.h"

/* maximum length of a line in a configuration file */
#define MAX_LINE_LEN 512


/* 
 * account_new()
 * 
 * Create a new account_t. Must be freed with account_free().
 */

account_t *account_new(void)
{
    account_t *a;

    a = (account_t *)xmalloc(sizeof(account_t));
    a->id = NULL;
    a->host = NULL;
    a->port = -1;
    a->from = NULL;
    a->auth_mech = NULL;
    a->username = NULL;
    a->password = NULL;
    a->tls = -1;
    a->tls_key_file = NULL;
    a->tls_cert_file = NULL;
    a->tls_trust_file = NULL;
    a->nostarttls = -1;
    a->keepbcc = -1;

    return a;
}


/* 
 * account_free()
 * 
 * Free an account_t.
 */

void account_free(account_t *a)
{
    if (a)
    {
	free(a->id);
	free(a->host);
	free(a->from);
	free(a->auth_mech);
	free(a->username);
	free(a->password);
	free(a->tls_key_file);
	free(a->tls_cert_file);
	free(a->tls_trust_file);
    }
}

void _account_free(void *p)
{
    account_free((account_t *)p);
}


/*
 * find_account()
 * 
 * Find an account in a list of accounts by its id.
 */

account_t *find_account(list_t *acc_list, char *id)
{
    list_t *p;
    account_t *a;
    
    p = acc_list->next;
    a = NULL;
    
    while (p->next != p)
    {
	if (strcmp(id, ((account_t *)(p->data))->id) == 0)
	{
	    a = (account_t *)p->data;
	    break;
	}
	p = p->next;
    }

    return a;
}

	    
/*
 * get_next_cmd()
 * 
 * Read a line from 'f'. Split it in a command part (first word after
 * whitespace) and an argument part (the word after the command).
 * Whitespace is ignored.
 * Return value:
 * 0 = ok
 * 1 = read error
 * 2 = empty line
 * -1 = EOF
 */

int skip_blanks(char *s, int i)
{
    while (s[i] != '\0' && (s[i] == ' ' || s[i] == '\t'))
    {
	i++;
    }
    return i;
}

int get_length(char *s)
{
    int i = 0;
    
    while (s[i] != '\0' && s[i] != ' ' && s[i] != '\t')
    {
	i++;
    }

    return i;
}

/* get last non-blank character in a line containing at least one */
int get_last_nonblank(char *s)
{
    int i;

    i = (int)strlen(s) - 1;
    while (s[i] == ' ' || s[i] == '\t')
    {
	i--;
    }
    return i;
}


int get_next_cmd(FILE *f, char **cmd, char **arg)
{
    char line[MAX_LINE_LEN];
    char *p;
    int i;
    int l;
    
    if (!fgets(line, (int)sizeof(line), f))
    {
	if (ferror(f))
	{
	    add_error_line("input error");
	    return 1;
	}
	else /* EOF */
	{
	    return -1;
	}
    }
    
    p = strchr(line, '\n');
    if (!p)
    {
	add_error_line("line longer than %d characters", MAX_LINE_LEN - 1);
	return 1;
    }

    *p = '\0';
    i = skip_blanks(line, 0);
    
    if (line[i] == '#' || line[i] == '\0')
    {
	return 2;
    }

    l = get_length(line + i);
    *cmd = (char *)xmalloc((l + 1) * sizeof(char));
    strncpy(*cmd, line + i, l);
    (*cmd)[l] = '\0';
    
    if (line[i])
    {
	i = skip_blanks(line, i + l + 1);
	l = get_last_nonblank(line + i);
	if (line[i] == '"' && line[i + l] == '"' && l > 0)
	{
    	    *arg = (char *)xmalloc(l * sizeof(char));
	    strncpy(*arg, line + i + 1, l - 1);
	    (*arg)[l] = '\0';
	}
	else
	{
	    *arg = (char *)xmalloc((l + 2) * sizeof(char));
	    strncpy(*arg, line + i, l + 1);
	    (*arg)[l + 2] = '\0';
	}
    }
    else
    {
	*arg = xstrdup("");
    }
    
    return 0;
}


/* 
 * read_conffile()
 * 
 * Read configuration data from 'f' and store it in 'acc_list'.
 * 'acc_list' will always be a new list; it may be empty if no accounts were
 * found.
 * Return values:
 * 0 = ok
 * 1 = read error
 * 2 = syntax error
 */

int read_conffile(FILE *f, list_t **acc_list)
{
    list_t *p;
    account_t *acc;
    char *cmd;
    char *arg;
    int error_code;
    int line;
    

    *acc_list = list_new();
    p = *acc_list;
    acc = NULL;
    error_code = 0;
    
    for (line = 1; ; line++)
    {
	if ((error_code = get_next_cmd(f, &cmd, &arg)) != 0)
	{
	    if (error_code == 1)
	    {
		break;
	    }
	    else if (error_code == 2)
	    {
		continue;
	    }
	    else /* error_code == -1, EOF */
	    {
		error_code = 0;
		break;
	    }
	}

	if (!acc && strcmp(cmd, "account") != 0)
	{
	    acc = account_new();
	    acc->id = xstrdup("default");
    	    list_insert(p, acc);
	    p = p->next;
	}
	if (strcmp(cmd, "account") == 0)
	{
	    if (find_account(*acc_list, arg))
	    {
		add_error_line("line %d: account `%s' was already defined", 
			line, arg);
		error_code = 2;
		break;
	    }
	    else
	    {
		acc = account_new();
		acc->id = xstrdup(arg);
		list_insert(p, acc);
		p = p->next;
	    }
	}
	else if (strcmp(cmd, "host") == 0)
	{
	    free(acc->host);
	    acc->host = xstrdup(arg);		
	}
	else if (strcmp(cmd, "port") == 0)
	{
	    acc->port = atoi(arg);
	    if (acc->port < 1 || acc->port > 65535)
	    {
		add_error_line("line %d: invalid port number %d", 
			line, acc->port);
		error_code = 2;
		break;
	    }
	}
	else if (strcmp(cmd, "from") == 0)
	{
	    free(acc->from);
	    acc->from = xstrdup(arg);
	}
	else if (strcmp(cmd, "auth") == 0)
	{
	    if (strcmp(arg, "") == 0)
	    {
		/* choose method later */
	    }
#ifdef USE_GSASL
	    else if (strcmp(arg, "digest-md5") == 0)
	    {
		acc->auth_mech = xstrdup("DIGEST-MD5");
	    }
	    else if (strcmp(arg, "ntlm") == 0)
	    {
		acc->auth_mech = xstrdup("NTLM");
	    }
#endif
	    else if (strcasecmp(arg, "plain") == 0)
	    {
		acc->auth_mech = xstrdup("PLAIN");
	    }
	    else if (strcasecmp(arg, "login") == 0)
	    {
		acc->auth_mech = xstrdup("LOGIN");
	    }
	    else if (strcasecmp(arg, "cram-md5") == 0)
	    {
		acc->auth_mech = xstrdup("CRAM-MD5");
	    }
	    else
	    {
		add_error_line("line %d: unknown authentication method `%s'", 
			line, arg);
		error_code = 2;
		break;
	    }
	}
	else if (strcmp(cmd, "user") == 0)
	{
	    free(acc->username);
	    acc->username = xstrdup(arg);
	}
	else if (strcmp(cmd, "password") == 0)
	{
	    free(acc->password);
	    acc->password = xstrdup(arg);
	}
	else if (strcmp(cmd, "tls") == 0)
	{
	    if (strcmp(arg, "") == 0)
	    {
		acc->tls = 1;
	    }
	    else
	    {
		add_error_line("line %d: command `tls' does not "
			"take an argument", line);
		error_code = 2;
		break;
	    }
	}
	else if (strcmp(cmd, "tls_key_file") == 0)
	{
	    free(acc->tls_key_file);
	    acc->tls_key_file = xstrdup(arg);		
	}
	else if (strcmp(cmd, "tls_cert_file") == 0)
	{
	    free(acc->tls_cert_file);
	    acc->tls_cert_file = xstrdup(arg);		
	}
	else if (strcmp(cmd, "tls_trust_file") == 0)
	{
	    free(acc->tls_trust_file);
	    acc->tls_trust_file = xstrdup(arg);		
	}
	else if (strcmp(cmd, "nostarttls") == 0)
	{
	    if (strcmp(arg, "") == 0)
	    {
		acc->nostarttls = 1;
	    }
	    else
	    {
		add_error_line("line %d: command `nostarttls' does not "
			"take an argument", line);
		error_code = 2;
		break;
	    }
	}
	else if (strcmp(cmd, "keepbcc") == 0)
	{
	    if (strcmp(arg, "") == 0)
	    {
		acc->keepbcc = 1;
	    }
	    else
	    {
		add_error_line("line %d: command `keepbcc' does not take an "
			"argument", line);
		error_code = 2;
		break;
	    }
	}
	else
	{
	    add_error_line("line %d: unknown command `%s'", line, cmd);
	    error_code = 2;
	    break;
	}    
    }
    
    return error_code;
}


/* 
 * complete_account()
 * 
 * Complete an account_t (fill in default values for fields not specified in the
 * configuration file).
 */

void complete_account(account_t *acc)
{
    if (acc->port == -1)
    {
	acc->port = 25;
    }
    if (acc->tls == -1)
    {
	acc->tls = 0;
    }
    if (!acc->auth_mech)
    {
	if (acc->username || acc->password)
	{
	    /* choose later */
    	    acc->auth_mech = xstrdup("");
	}
	else
	{
	    /* NULL means no authentication */
	}
    }
    if (acc->nostarttls == -1)
    {
	acc->nostarttls = 0;
    }
    if (acc->keepbcc == -1)
    {
	acc->keepbcc = 0;
    }	
}


/*
 * check_account()
 * 
 * Check an account_t. 0 = ok, 1 = error.
 */

int check_account(account_t *acc)
{
    if (!acc->id)
    {
	add_error_line("account has no name");
	return 1;
    }
    if (!acc->host)
    {
	add_error_line("account `%s' has no host information", acc->id);
	return 1;
    }
    if (!acc->from)
    {
	add_error_line("account `%s' has no envelope from information", acc->id);
	return 1;
    }
    if (acc->auth_mech && (!acc->username || !acc->password))
    {
	add_error_line("account `%s' needs both username and password "
		"for authentication", acc->id);
	return 1;
    }
    if (acc->tls_key_file && !acc->tls)
    {
	add_error_line("account `%s' has the `tls_key_file' command but does not use TLS",
		acc->id);
	return 1;
    }
    if (acc->tls_cert_file && !acc->tls)
    {
	add_error_line("account `%s' has the `tls_cert_file' command but does not use TLS",
		acc->id);
	return 1;
    }
    if (acc->tls_key_file && !acc->tls_cert_file) 
    {
	add_error_line("account `%s' has the `tls_key_file' command but not the `tls_cert_file' command",
		acc->id);
	return 1;
    }
    if (!acc->tls_key_file && acc->tls_cert_file)
    {
	add_error_line("account `%s' has the `tls_cert_file' command but not the `tls_key_file' command",
		acc->id);
	return 1;
    }
    if (acc->tls_trust_file && !acc->tls)
    {
	add_error_line("account `%s' has the `tls_trust_file' command but does not use TLS",
		acc->id);
	return 1;
    }
    if (acc->nostarttls && !acc->tls)
    {
	add_error_line("account `%s' has the `nostarttls' command but does not use TLS",
		acc->id);
	return 1;
    }

    return 0;
}

	
/*
 * get_conf()
 * 
 * Read 'conffile' and store all account data in 'acc_list'.
 * Return values:
 * 0 = ok
 * 1 = io error
 * 2 = syntax error
 * 3 = invalid account(s)
 */

int get_conf(char *conffile, list_t **acc_list)
{
    FILE *f;
    int error_code;
    list_t *p;

    if (!(f = fopen(conffile, "r")))
    {
	add_error_line("cannot open file `%s' [%s]", conffile, strerror(errno));
	return 1;
    }

    error_code = read_conffile(f, acc_list);
    fclose(f);
    if (error_code != 0)
    {
	if (error_code == 1)
	{
	    add_error_line("input error in `%s'", conffile);
	}
	else /* error_code == 2 */
	{
	    add_error_line("syntax error in `%s'", conffile);
	}
	list_xfree(*acc_list, _account_free);
	return error_code;
    }
    
    if (list_is_empty(*acc_list))
    {
	add_error_line("no valid configuration data in `%s'", conffile);
	list_free(*acc_list);
	return 3;
    }

    p = (*acc_list)->next;
    while (p->next != p)
    {
	complete_account((account_t *)(p->data));
	if (check_account((account_t *)(p->data)) != 0)
	{
	    add_error_line("invalid account in `%s'", conffile);
	    list_xfree(*acc_list, _account_free);
	    return 3;
	}
	p = p->next;
    }

    return 0;
}
