/*
 * net.c
 * 
 * This file is part of the msmtp package, a very simple programm that 
 * works as an "SMTP plugin" for Mutt and possibly other MUAs.
 *
 * Copyright (C) 2000, 2003	Martin Lambers <marlam@web.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <unistd.h>
#include <string.h>
#include <errno.h>
extern int errno;
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>

#include "net.h"
#include "merror.h"
#include "xmalloc.h"


/*
 * open_socket() 
 *
 * Opens a TCP socket to 'hostname':'port'.
 * 'hostname' may be a host name, an IPv4 address in dot notation, or an IPv6 
 * address in colon notation.
 * On success, the file descriptor is returned. 
 * On failure, -1 is returned.
 */

int open_socket(char *hostname, int port)
{
    int fd;
    char *port_string;
    struct addrinfo hints;
    struct addrinfo *res0;
    struct addrinfo *res;
    int error_code;
    int cause;
    
    hints.ai_family = PF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = 0;
    hints.ai_protocol = 0;
    hints.ai_addrlen = 0;
    hints.ai_canonname = NULL;
    hints.ai_addr = NULL;
    hints.ai_next = NULL;
    xasprintf(&port_string, "%d", port);
    error_code = getaddrinfo(hostname, port_string, &hints, &res0);
    free(port_string);
    if (error_code)
    {
	add_error_line("cannot locate host `%s' [%s]",
		hostname, gai_strerror(error_code));
	return -1;
    }

    fd = -1;
    cause = 0;
    for (res = res0; res; res = res->ai_next)
    {
	fd = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
	if (fd < 0)
	{
	    cause = 1;
	    continue;
	}
	if (connect(fd, res->ai_addr, res->ai_addrlen) < 0) 
	{
	    cause = 2;
	    close(fd);
	    fd = -1;
	    continue;
	}
	break;
    }
    freeaddrinfo(res0);
    if (fd < 0)
    {
	if (cause == 1)
	{
	    add_error_line("cannot create socket [socket(): %s]", 
		    strerror(errno));
	}
	else if (cause == 2)
	{
	    add_error_line("cannot connect to `%s', port %d [connect(): %s]", 
		    hostname, port, strerror(errno));
	}
	else /* cause == 0, can't happen */
	{
	    add_error_line("cannot locate host `%s' "
		    "[getaddrinfo() returned crap]", hostname);
	}
	return -1;
    }
    
    return fd;
}


/*
 * net_getchar()
 *
 * Reads a character from 'fd' and stores it in 'c'.
 * Return value is 0 on success, -1 on EOF and 1 on error.
 */

int net_getchar(int fd, char *c)
{
    ssize_t error_code;
    
    if ((error_code = read(fd, c, (size_t)1)) == 1)
    {
	return 0;
    }
    else if (error_code == 0)
    {
	return -1;
    }
    else /* error_code == -1 */
    {
	add_error_line("read error [read(): %s]", strerror(errno));
	return 1;
    }
}


/*
 * net_getline()
 * 
 * Reads a line from 'fd' into 'line', which is of size 'size'.
 * The line end (either '\n' or '\r\n') is not stored in 'line'. 
 * 'line' will alway be terminated by '\0'.
 * Return value is 0 on success, 1 on read error and 2 if the line is longer
 * than 'size - 1' characters.
 */

int net_getline(int fd, char *line, int size)
{
    char c;
    int i;
    int error_code;

    i = 0;
    while ((error_code = net_getchar(fd, &c)) == 0)
    {
	if (i == size - 1)
    	{
	    /* line too long */
	    line[i] = '\0';
	    return 2;
	}
	else if (c == '\n')
	{
	    break;
	}
	else
	{
	    line[i++] = c;
	}
    }
    if (c == '\n' && line[i - 1] == '\r')
    {
	line[i - 1] = '\0';
    }
    line[i] = '\0';
    if (error_code == 0 || error_code == -1)
    {
	return 0;
    }
    else
    {
	return 1;
    }
}


/*
 * net_putline()
 *
 * Writes a line to 'fd'.
 * Return value is 0 on success and 1 on write error.
 */

int net_putline(int fd, char *line)
{
    ssize_t error_code;
    size_t count;

    count = strlen(line);    
    error_code = write(fd, line, count);
    if (error_code < 0)
    {
	add_error_line("write error [write(): %s]", strerror(errno));
	return 1;
    }
    else if ((size_t)error_code == count)
    {
	return 0;
    }
    else /* 0 <= error_code < count */
    {
	add_error_line("write error [write(): unknown error]");
	return 1;
    }
}
