/*
 * tls.c
 *
 * This file is part of the msmtp package, a very simple programm that 
 * works as an "SMTP plugin" for Mutt and possibly other MUAs.
 *
 * Copyright (C) 2000, 2003	Martin Lambers <marlam@web.de>
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <string.h>
#include <limits.h>
#include <time.h>

#ifdef USE_GNUTLS
#include <gnutls/gnutls.h>
#else
#include <openssl/ssl.h>
#include <openssl/x509.h>
#include <openssl/err.h>
#include <openssl/rand.h>
#endif

#include "merror.h"
#include "tls.h"


/*
 * Always use tls_clear() or tls_init()/tls_start() before using a tls_t!
 * Never call a tls_*() function with tls_t NULL!
 */


/*
 * tls_clear()
 *
 * Clears a tls_t type (marks it inactive).
 */

void tls_clear(tls_t *tls)
{
#ifdef USE_GNUTLS
    tls->is_active = 0;
#else
    tls->ssl_ctx = NULL;
    tls->ssl = NULL;
#endif
}


/*
 * tls_init()
 *
 * Initialize underlying TLS library.
 * Returns 0 on success and 1 on error.
 */

int tls_init(void)
{
#ifdef USE_GNUTLS
    int error_code;
    
    if ((error_code = gnutls_global_init()) != 0)
    {
	add_error_line("Cannot initialize GnuTLS library [%s]", 
		gnutls_strerror(error_code));
	return 1;
    }
    return 0;
    
#else

    SSL_load_error_strings();
    SSL_library_init();
    if (!RAND_status)
    {
	/* FIXME: seed random number generator 
	 * On systems whith /dev/urandom, OpenSSL seeds itself.
	 * On other systems, we have to do it ourselves with RAND_seed().
	 * I have no idea with what. */
	add_error_line("BUG: must seed random number generator for TLS [%s]", 
		ERR_error_string(ERR_get_error(), NULL));
	return 1;
    }

    return 0;
#endif
}


/*
 * tls_is_active()
 *
 * Returns 1 if 'tls' is an active TLS connection, otherwise 0.
 */

int tls_is_active(tls_t *tls)
{
#ifdef USE_GNUTLS
    return tls->is_active;
#else
    return (tls->ssl != NULL);
#endif
}


/*
 * openssl_io_error()
 *
 * Used only internally by the OpenSSL code.
 * 
 * Add an error line according to 'error_code' (which originates from an
 * SSL_read(), SSL_write() or SSL_connect() operation) and 'error_code2' (which
 * originates from an SSL_get_error() call with 'error_code' as its argument).
 * The line will read: "error_string [error_reason]". 'error_string' is given by
 * the calling function, this function finds out 'error_reason'.
 */

#ifndef USE_GNUTLS
void openssl_io_error(int error_code, int error_code2, char *error_string)
{
    unsigned long error_code3;
    char *error_reason;
    
    switch (error_code2)
    {
	case SSL_ERROR_NONE:
	    error_reason = "unknown error";
	    break;
		
	case SSL_ERROR_SYSCALL:
	    error_code3 = ERR_get_error();
    	    if (error_code3 == 0)
	    {
		if (error_code == 0)
		{
		    error_reason = "a protocol violating EOF occured";
		}
		else if (error_code == -1)
		{
		    error_reason = strerror(errno);
		}
		else
		{
		    error_reason = "unknown error";
		}
	    }
	    else
	    {
		error_reason = ERR_error_string(error_code3, NULL);
	    }
	    break;
	    
	case SSL_ERROR_ZERO_RETURN:
	    error_reason = "the connection has been closed unexpectedly";
	    break;

	case SSL_ERROR_SSL:
	    error_reason = ERR_error_string(ERR_get_error(), NULL);
	    break;
	    
	default:
    	    error_reason = "unknown error";
	    break;
    }
    add_error_line("%s [%s]", error_string, error_reason);
}
#endif


/* 
 * tls_check_cert()
 *
 * If the 'verify' flag is set, perform a real verification of the peer's 
 * certificate. If this succeeds, the connection can be considered secure.
 * If the 'verify' flag is not set, perform only a few sanity checks of the
 * peer's certificate. You cannot trust the connection when this succeeds.
 * Returns 0 on success, 1 otherwise.
 */

int tls_check_cert(tls_t *tls, char *hostname, int verify)
{
#ifdef USE_GNUTLS
    int status;
    char *error_msg;
    const gnutls_datum* cert_list;
    int cert_list_size;
    
    if (verify)
    {
	error_msg = "TLS certificate verification failed [%s]";
    }
    else
    {
	error_msg = "TLS certificate check failed [%s]";
    }
    
    /* This verification function uses the trusted CAs in the credentials
     * structure. So you must have installed one or more CA certificates.
     */
    status = gnutls_certificate_verify_peers(tls->session);
    if (status == GNUTLS_E_NO_CERTIFICATE_FOUND) 
    {
	add_error_line(error_msg, "No certificate was sent");
	return 1;
    }
    if (status & GNUTLS_CERT_INVALID || status & GNUTLS_CERT_CORRUPTED)
    {
	add_error_line(error_msg, "The certificate is invalid");
	return 1;
    }
    if (verify)
    {
	if (status & GNUTLS_CERT_NOT_TRUSTED || status & GNUTLS_CERT_REVOKED)
	{
	    add_error_line(error_msg, "The certificate is not trusted");
	    return 1;
	}
    }
    if (gnutls_certificate_expiration_time_peers(tls->session) < time(NULL)) 
    {
	add_error_line(error_msg, "The certificate has expired");
	return 1;
    }
    if (gnutls_certificate_activation_time_peers(tls->session) > time(NULL)) 
    {
       	add_error_line(error_msg, "The certificate is not yet activated");
	return 1;
    }
    if (gnutls_certificate_type_get(tls->session) == GNUTLS_CRT_X509) 
    {
	cert_list = gnutls_certificate_get_peers(tls->session, &cert_list_size);
	if (cert_list == NULL) 
	{
	    add_error_line(error_msg, "No certificate was found!");
	    return 1;
	}
	if (!gnutls_x509_check_certificates_hostname(&cert_list[0], hostname))
	{
	    add_error_line(error_msg, "The certificate does not match the hostname");
	    return 1;
	}
    }
    return 0;

#else

    X509 *x509cert;
    X509_NAME *x509_subject;
    long status;
    char buf[257];
    size_t l;
    char *s1, *s2;
    char *error_msg;
    

    if (verify)
    {
	error_msg = "TLS certificate verification failed [%s]";
    }
    else
    {
	error_msg = "TLS certificate check failed [%s]";
    }

    /* Get certificate */
    if (!(x509cert = SSL_get_peer_certificate(tls->ssl)))
    {
	add_error_line(error_msg, "No certificate was sent");
	return 1;
    }
    
    /* Call OpenSSL's default verify function */
    if ((status = SSL_get_verify_result(tls->ssl)) != X509_V_OK)
    {
	if (!verify && status == X509_V_ERR_UNABLE_TO_GET_ISSUER_CERT_LOCALLY)
	{
	    X509_free(x509cert);
	    return 0;
	}
	else
	{
    	    add_error_line(error_msg, X509_verify_cert_error_string(status));
	    X509_free(x509cert);
	    return 1;
	}
    }

    /* Check if hostname matches certificate Common Name (CN) */
    if (!(x509_subject = X509_get_subject_name(x509cert)))
    {
	add_error_line(error_msg, "cannot get certificate subject");
	X509_free(x509cert);
	return 1;
    }
    if (X509_NAME_get_text_by_NID(
		x509_subject, NID_commonName, buf, sizeof(buf)) == -1)
    {
	add_error_line(error_msg, "cannot get certificate common name");
	X509_free(x509cert);
	return 1;
    }
    /* left-most character may be wildcard */
    if (buf[0] == '*')
    {
	s1 = buf + 1;
	l = strlen(s1);
	s2 = hostname + strlen(hostname) - l;
    }
    else
    {
	s1 = buf;
	s2 = hostname;
    }
    if (strcasecmp(buf, hostname) != 0)
    {
	add_error_line(error_msg, 
		"The certificate does not match the hostname");
	X509_free(x509cert);
	return 1;
    }
    X509_free(x509cert);
    return 0;
#endif
}


/*
 * tls_start
 *
 * Starts TLS encryption on a socket.
 * If 'key_file' and 'cert_file' are not NULL, they contain a client certificate
 * information to be sent to the server if requested.
 * If 'ca_file' is NULL, only sanity checks are performed on the peer
 * certificate. If 'ca_file' is a PEM file containing the certificate of a 
 * Certification Authority, real verification of the peer certificate is performed.
 * 'hostname' is the host to start TLS with. It is needed for sanity checks/
 * verification.
 * Returns 0 on success and 1 on error.
 */

int tls_start(tls_t *tls, int fd, char *hostname, 
	char *key_file, char *cert_file, char *ca_file)
{
#ifdef USE_GNUTLS
    int error_code;
    
    if ((error_code = gnutls_init(&tls->session, GNUTLS_CLIENT)) != 0)
    {
	add_error_line("Cannot initialize TLS Session [%s]",
		gnutls_strerror(error_code));
	return 1;
    }
    if ((error_code = gnutls_set_default_priority(tls->session)) != 0)
    {
	add_error_line("Cannot set priorities on TLS Session [%s]",
		gnutls_strerror(error_code));
	gnutls_deinit(tls->session);
	return 1;
    }
    if ((error_code = gnutls_certificate_allocate_credentials(&tls->cred)) < 0)
    {
	add_error_line("Cannot allocate certificate for TLS Session [%s]",
		gnutls_strerror(error_code));
	gnutls_deinit(tls->session);
	return 1;
    }
    if (key_file && cert_file)
    {
	if ((error_code = gnutls_certificate_set_x509_key_file(tls->cred, 
		       	cert_file, key_file, GNUTLS_X509_FMT_PEM)) < 0)
	{
	    add_error_line("Cannot set X509 key file `%s' and/or "
		    "X509 cert file `%s' for TLS Session [%s]",
	    	    key_file, cert_file, gnutls_strerror(error_code));
	    gnutls_deinit(tls->session);
	    gnutls_certificate_free_credentials(tls->cred);
	    return 1;
	}
    }
    if (ca_file)
    {
	if ((error_code = gnutls_certificate_set_x509_trust_file(tls->cred, 
			ca_file, GNUTLS_X509_FMT_PEM)) <= 0)
	{
	    add_error_line("Cannot set X509 trust file `%s' for TLS Session [%s]",
	    	    ca_file, gnutls_strerror(error_code));
	    gnutls_deinit(tls->session);
	    gnutls_certificate_free_credentials(tls->cred);
	    return 1;
	}
    }
    if ((error_code = gnutls_credentials_set(tls->session, 
		    GNUTLS_CRD_CERTIFICATE, tls->cred)) < 0)
    {
	add_error_line("Cannot set credentials for TLS Session [%s]",
		gnutls_strerror(error_code));
	gnutls_deinit(tls->session);
	gnutls_certificate_free_credentials(tls->cred);
	return 1;
    }
    gnutls_transport_set_ptr(tls->session, (gnutls_transport_ptr)fd);
    if ((error_code = gnutls_handshake(tls->session)) < 0)
    {
	gnutls_deinit(tls->session);
	gnutls_certificate_free_credentials(tls->cred);
	add_error_line("TLS handshake failed [%s]", gnutls_strerror(error_code));
	return 1;
    }
    if (tls_check_cert(tls, hostname, ca_file ? 1 : 0) != 0)
    {
	gnutls_deinit(tls->session);
	gnutls_certificate_free_credentials(tls->cred);
	return 1;
    }
    
    tls->is_active = 1;
    return 0;
    
#else
    
    SSL_METHOD *ssl_method = NULL;
    int error_code;
    
    if (!(ssl_method = SSLv23_client_method()))
    {
	add_error_line("SSLv23_client_method() failed");
	return 1;
    }
    tls->ssl_ctx = SSL_CTX_new(ssl_method);
    if (!tls->ssl_ctx)
    {
	add_error_line("cannot create TLS context [%s]", 
		ERR_error_string(ERR_get_error(), NULL));
	return 1;
    }
    if (key_file && cert_file)
    {
	if (SSL_CTX_use_PrivateKey_file(tls->ssl_ctx, key_file, SSL_FILETYPE_PEM) != 1)
	{
	    add_error_line("cannot load key file `%s' [%s]", 
		    key_file, ERR_error_string(ERR_get_error(), NULL));
	    SSL_CTX_free(tls->ssl_ctx);
	    tls->ssl_ctx = NULL;
	    return 1;
	}
	if (SSL_CTX_use_certificate_chain_file(tls->ssl_ctx, cert_file) != 1)
	{
	    add_error_line("cannot load certification file `%s' [%s]", 
		    cert_file, ERR_error_string(ERR_get_error(), NULL));
	    SSL_CTX_free(tls->ssl_ctx);
	    tls->ssl_ctx = NULL;
	    return 1;
	}
    }
    if (ca_file)
    {
	if (SSL_CTX_load_verify_locations(tls->ssl_ctx, ca_file, NULL) != 1)
	{
	    add_error_line("cannot load CA file `%s' [%s]", 
		    ca_file, ERR_error_string(ERR_get_error(), NULL));
	    SSL_CTX_free(tls->ssl_ctx);
	    tls->ssl_ctx = NULL;
	    return 1;
	}
    }
    if (!(tls->ssl = SSL_new(tls->ssl_ctx)))
    {
	add_error_line("SSL_new() failed [%s]", 
		ERR_error_string(ERR_get_error(), NULL));
	SSL_CTX_free(tls->ssl_ctx);
	tls->ssl_ctx = NULL;
	return 1;
    }
    if (!SSL_set_fd(tls->ssl, fd))
    {
	add_error_line("SSL_set_fd() failed [%s]", 
		ERR_error_string(ERR_get_error(), NULL));
	SSL_free(tls->ssl);
	SSL_CTX_free(tls->ssl_ctx);
	tls->ssl = NULL;
	tls->ssl_ctx = NULL;
	return 1;
    }
    if ((error_code = SSL_connect(tls->ssl)) < 1)
    {
	openssl_io_error(error_code, SSL_get_error(tls->ssl, error_code),
		"cannot establish TLS connection");
	SSL_free(tls->ssl);
	SSL_CTX_free(tls->ssl_ctx);
	tls->ssl = NULL;
	tls->ssl_ctx = NULL;
	return 1;
    }
    if (tls_check_cert(tls, hostname, ca_file ? 1 : 0) != 0)
    {
	SSL_free(tls->ssl);
	SSL_CTX_free(tls->ssl_ctx);
	tls->ssl = NULL;
	tls->ssl_ctx = NULL;
	return 1;
    }

    return 0;
#endif
}


/*
 * tls_getchar()
 *
 * Reads a character using TLS and stores it in 'c'.
 * Returns 0 on success, -1 on EOF and 1 on error.
 */

int tls_getchar(tls_t *tls, char *c)
{
#ifdef USE_GNUTLS
    ssize_t error_code;
    
    error_code = gnutls_record_recv(tls->session, c, 1);
    if (error_code == 1)
    {
	return 0;
    }
    else if (error_code == 0)
    {
	return -1;
    }
    else /* error_code < 0 */
    {
	add_error_line("cannot read from TLS connection [%s]", 
		gnutls_strerror(error_code));
	return 1;
    }
    
#else
    
    int error_code;
    int error_code2;
    
    if ((error_code = SSL_read(tls->ssl, c, 1)) < 1)
    {
	if ((error_code2 = SSL_get_error(tls->ssl, error_code)) 
		== SSL_ERROR_NONE)
	{
	    /* EOF */
	    return -1;
	}
	else
	{
	    openssl_io_error(error_code, error_code2, 
		    "cannot read from TLS connection");
	    return 1;
	}
    }
    else
    {
	return 0;
    }
#endif
}


/*
 * tls_getline()
 *
 * Reads a line from 'tls' into 'line', which is of size 'size'.
 * The line end (either '\n' or '\r\n') is not stored in 'line'. 
 * 'line' will alway be terminated by '\0'.
 * Return value is 0 on success, 1 on read error and 2 if the line is longer
 * than 'size - 1' characters.
 */

int tls_getline(tls_t *tls, char *line, int size)
{
    char c;
    int i;
    int error_code;

    i = 0;
    while ((error_code = tls_getchar(tls, &c)) == 0)
    {
	if (i == size - 1)
    	{
	    /* line too long */
	    line[i] = '\0';
	    return 2;
	}
	else if (c == '\n')
	{
	    break;
	}
	else
	{
	    line[i++] = c;
	}
    }
    if (c == '\n' && line[i - 1] == '\r')
    {
	line[i - 1] = '\0';
    }
    line[i] = '\0';
    if (error_code == 0 || error_code == -1)
    {
	return 0;
    }
    else
    {
	return 1;
    }
}


/*
 * tls_putline()
 *
 * Writes a line using TLS.
 * Returns 0 on success and 1 on error.
 */

int tls_putline(tls_t *tls, char *line)
{
#ifdef USE_GNUTLS
    ssize_t error_code;
    size_t count;

    count = strlen(line);    
    error_code = gnutls_record_send(tls->session, line, count);
    if (error_code < 0)
    {
	add_error_line("cannot write to TLS connection [%s]", 
		gnutls_strerror(error_code));
	return 1;
    }
    else if ((size_t)error_code == count)
    {
	return 0;
    }
    else /* 0 <= error_code < count */
    {
	add_error_line("cannot write to TLS connection [unknown error]");
	return 1;
    }    

#else
    
    int error_code;
    size_t count;

    count = strlen(line);
    if (count > (size_t)INT_MAX)
    {
	add_error_line("line too long in tls_putline()");
	return 1;
    }
    else if (count < 1)
    {
	/* nothing to be done */
	return 0;
    }
    
    if ((error_code = SSL_write(tls->ssl, line, (int)count)) != (int)count)
    {
	openssl_io_error(error_code, SSL_get_error(tls->ssl, error_code),
		"cannot write to TLS connection");
	return 1;
    }

    return 0;
#endif
}


/*
 * tls_close()
 *
 * Close a TLS connection and mark it inactive
 */

void tls_close(tls_t *tls)
{
#ifdef USE_GNUTLS
    if (tls->is_active)
    {
	gnutls_bye(tls->session, GNUTLS_SHUT_RDWR);
	gnutls_deinit(tls->session);
	gnutls_certificate_free_credentials(tls->cred);
	tls->is_active = 0;
    }
    
#else
    
    if (tls->ssl)
    {
 	SSL_shutdown(tls->ssl);
	SSL_free(tls->ssl);
    }
    if (tls->ssl_ctx)
    {
	SSL_CTX_free(tls->ssl_ctx);
    }
#endif
}


/*
 * tls_deinit()
 *
 * Deinit underlying TLS library.
 */

void tls_deinit(void)
{
#ifdef USE_GNUTLS
    gnutls_global_deinit();
#endif
}
